from PyQt5 import uic
from PyQt5.QtWidgets import QFrame

from .io import find_ui
from pydevmgr_core import NodeVar
from .vltdevice_ctrl import VltDeviceCtrl
from pydevmgr_core_qt import record_widget_factory

from enum import IntEnum

#####
# Define all the data used at fron-end 
# The NodeVar annotation indicate that the data will be taken from a device node see `pydevmgr.DataLink`

class VltMotorCtrlStatData(VltDeviceCtrl.Data.StatData):
    # DeviceStatData contain state, substate, error code and txt information 
    # Add here all NodeVar needed for the motor GUI    
    pos_actual: NodeVar[float] = 0.0
    pos_target: NodeVar[float] = 0.0
    vel_actual: NodeVar[float] = 0.0

###
# Example of use:
#  data = MotorCtrlData()
#  dl = DataLink(motor, data)
#  dl.update() # update data from server(s)
#  print(data.stat.pos_actual, data.stat.pos_error) 
class VltMotorCtrlData(VltDeviceCtrl.Data):
    StatData = VltMotorCtrlStatData
    stat: StatData = StatData()

class MOVE_MODE(IntEnum):
    ABSOLUTE = 0
    RELATIVE = 1
    VELOCITY = 2

# The UI is an empty shell all inteligence is settled by MotorCtrlLink
class VltMotorCtrlUi(QFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        uic.loadUi(find_ui('vltmotor_ctrl_frame.ui'), self)

    
class VltMotorCtrl(VltDeviceCtrl):
    # A widget constructor for default widget 
    Widget = VltMotorCtrlUi
    # A data constructor 
    Data = VltMotorCtrlData
    
    # Basicaly 3 methods needs to be implemented 
    # - .init_vars() :  build the variable handler function to their type and widget
    # - .update(data) :   update widget with new data 
    # - .connect_device(device, data) : this is where commands has to be connected to widget buttons etc...
    
    def init_vars(self):
        super().init_vars()
        
        # Add an interface between the widget and the variable to set/get 
        # the goal is to give basic functionality to the variable (.get ) wihtout having to care 
        # on the widget type. It is expected that the .Float will handled any suitable widgets
        # however so far only a few widget are handled and this can involved function to usage
        # It is okay to also use the widget directly, the goal of these layer is not to interface all
        # widget capabilities but only some basic ones.
        # outputs/inputs are instances of Inputs/Outputs and can be use to collect constructors
        # and instances of widget interfaces. 
        self.outputs.pos_actual = self.outputs.Float(self.widget.pos_actual, fmt="%.3f")
        self.outputs.pos_target = self.outputs.Float(self.widget.pos_target, fmt="%.3f")
        self.outputs.vel_actual = self.outputs.Float(self.widget.vel_actual, fmt="%.3f")
        
        self.outputs.pos_error = self.outputs.Float(self.widget.pos_error, fmt="%.3f")
        self.outputs.state = self.outputs.Code(self.widget.state)

        self.inputs.velocity   = self.inputs.Float(self.widget.input_velocity, default=1.0)
        self.inputs.pos_target = self.inputs.Float(self.widget.input_pos_target)
        self.inputs.move_mode  = self.inputs.Enum(MOVE_MODE , self.widget.move_mode, default=MOVE_MODE.ABSOLUTE)     
     
    def update(self, data):
        super().update(data)
        # data as returned by self.new_data()
        stat = data.stat 
        
        if self.widget.move_mode.currentIndex()==MOVE_MODE.VELOCITY: # VELOCITY
            self.widget.input_pos_target.setEnabled(False)
        else:
            self.widget.input_pos_target.setEnabled(True)
        
        # this will update the widget labels underneath          
        self.outputs.state.set( (stat.state, stat.state_txt ) )
        self.outputs.pos_actual.set(stat.pos_actual)
        self.outputs.pos_error.set(stat.pos_actual - stat.pos_target)
        self.outputs.pos_target.set(stat.pos_target)
        self.outputs.vel_actual.set(stat.vel_actual)
    
    
    def setup_ui(self,motor, data):
        # super shall take care of state and substate         
        super().setup_ui(motor, data)
                                
        # init some input fields to default, only if the device is connected         
        if motor.is_connected():
            self.inputs.velocity.set_input(motor.cfg.velocity.get())
        
        # self.actions is an instance of Actions and shall be used to implement any action
        # it is important to use actions.add instead of directly Action() as some individual action 
        # need a proper global setup when connected to a widget (e.g. connect_item)
        # The possibilities of action connection type is still reduced but should grows in the future
                
        # Stop button calls motor.stop without argument feedback(er, txt) is called in case of exception            
        self.actions.add( motor.stop, feedback=self.feedback).connect_button(self.widget.stop)
        
        # MOVE button 
        # define a move function to handled inputs comming from movement type, pos and velocity                               
        def move(mode, pos, vel):
            if mode == MOVE_MODE.ABSOLUTE:
                motor.move_abs(pos, vel)
            elif mode == MOVE_MODE.RELATIVE:
                motor.move_rel(pos, vel)
            elif mode == MOVE_MODE.VELOCITY:
                motor.move_vel(vel)
        
        move = self.actions.add(
                      move, 
                      # if a function (like bellow) it is called at run time to get the value
                      # if an exception occurs feedback is called and move function is not 
                      [self.inputs.move_mode.get, self.inputs.pos_target.get, self.inputs.velocity.get], 
                      feedback = self.feedback
                      )
        move.connect_button(self.widget.move)
    
    def setup_actions(self, device, data):
        super().setup_actions(device, data)
        

                                           
record_widget_factory("ctrl", "VltMotor", VltMotorCtrl)




from .vltdevice import VltDevice
from pydevmgr_core import record_class, BaseParser , upload,  NodeAlias, NodeAlias1, NodeVar
from ._tools import _inc, NegNode
from pydevmgr_ua import Int32, Int16
from pydantic import BaseModel , validator, root_validator
from typing import List, Dict, Any, Optional, Union 
from enum import Enum
from collections import OrderedDict



class VltMotorCtrlConfig(BaseModel):
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    velocity : float = 0.1 # mendatory because used as default for movement
    min_pos :           Optional[float] = 0.0
    max_pos :           Optional[float] = 0.0 
    axis_type :         Union[None,int,str] = "LINEAR" # LINEAR , CIRCULAR, CIRCULAR_OPTIMISED
    active_low_lstop :  Optional[bool] = False
    active_low_lhw :    Optional[bool] = False
    active_low_ref :    Optional[bool] = True
    active_low_index :  Optional[bool] = False
    active_low_uhw :    Optional[bool] = True
    active_low_ustop :  Optional[bool] = False
    brake :             Optional[bool] = False
    low_brake :         Optional[bool] = False
    backlash :          Optional[float] = 0.0
    tout_init :         Optional[int] = 30000
    tout_move :         Optional[int] = 12000
    tout_switch :       Optional[int] = 10000
    
    scale_factor : Optional[float] = 1.0
    accel: Optional[float] = 30.0
    decel: Optional[float] = 30.0
    jerk:  Optional[float] = 100.0
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Validator Functions
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
    @validator('axis_type')
    def validate_axis_type(cls, ax):
        if isinstance(ax, str):
            try:
                getattr(AXIS_TYPE, ax)
            except AttributeError:
                raise ValueError(f"Unknown axis_type {ax!r}")
        if isinstance(ax, int):            
            # always return a string??
            ax = AXIS_TYPE(ax).name        
        return ax

    
class PositionsConfig(BaseModel):
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    posnames : List = []
    tolerance: float = 1.0
    positions: Dict = OrderedDict()  # adding a dictionary for positions. Presfered than leaving it as extra 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Validator Functions
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class Config:        
        extra = 'allow' # needed for the poses 
        validate_assignment = True
    @root_validator()
    def collect_positions(cls, values):     
        """ collectect the positions from the extras """ 
        positions = values['positions']
        for name in values['posnames']:
            if name not in positions:
                try:
                    positions[name] = float( values[name] ) 
                except (KeyError, TypeError):
                    raise ValueError(f'posname {name!r} is not defined or not a float')   
        return values 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Method to save back the configuration     
    def cfgdict(self):
        d = {'posnames': self.posnames, 'tolerance':self.tolerance}
        for p in self.posnames:
            d[p] = self.positions[p]
        return d


# ################################  
class SeqStepConfig(BaseModel):
    """ Data  Model for step configuration """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    index: int =  0    
    value1: float = 0.0
    value2: float = 0.0
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~
    class Config:                
        validate_assignment = True
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Method to save back the configuration     
    def cfgdict(self):
        return self.dict(exclude={"index"})


class InitialisationConfig(BaseModel):
    """ Config Model for the initialisation sequence """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    sequence : List[str] = []
    END          : SeqStepConfig = SeqStepConfig(index=0)
    FIND_INDEX   : SeqStepConfig = SeqStepConfig(index=1)
    FIND_REF_LE  : SeqStepConfig = SeqStepConfig(index=2)
    FIND_REF_UE  : SeqStepConfig = SeqStepConfig(index=3)
    FIND_LHW     : SeqStepConfig = SeqStepConfig(index=4)
    FIND_UHW     : SeqStepConfig = SeqStepConfig(index=5)  
    DELAY        : SeqStepConfig = SeqStepConfig(index=6)
    MOVE_ABS     : SeqStepConfig = SeqStepConfig(index=7)
    MOVE_REL     : SeqStepConfig = SeqStepConfig(index=8)
    CALIB_ABS    : SeqStepConfig = SeqStepConfig(index=9)
    CALIB_REL    : SeqStepConfig = SeqStepConfig(index=10)
    CALIB_SWITCH : SeqStepConfig = SeqStepConfig(index=11)
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~
    class Config:                
        validate_assignment = True        
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Validator Functions
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @validator('END', 'FIND_INDEX', 'FIND_REF_LE', 'FIND_REF_UE', 'FIND_LHW', 'FIND_UHW', 
               'DELAY', 'MOVE_ABS', 'MOVE_REL', 'CALIB_ABS', 'CALIB_REL' , 'CALIB_SWITCH')
    def force_index(cls, v, field):
        """ need to write the index """        
        v.index = getattr(INITSEQ, field.name)
        return v

    @validator('sequence')
    def validate_initialisation(cls,sequence):   
        """ Validate the list of sequence """ 
        for s in sequence:
            try:
                cls.__fields__[s]
            except KeyError:
                raise ValueError(f'unknown sequence step named {s!r}')
        return sequence
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Method to save back the configuration     
    def cfgdict(self):
        d = {'sequence':self.sequence}
        for seq in self.sequence:
            d[seq] = getattr(self, seq).cfgdict()            
        return d
# ################################

class VltMotorConfig(VltDevice.Config):
    CtrlConfig = VltMotorCtrlConfig
    PositionsConfig = PositionsConfig
    InitialisationConfig = InitialisationConfig
    

    type = "VltMotor"
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    type: str = "Motor"
    initialisation : InitialisationConfig = InitialisationConfig()
    positions      : PositionsConfig = PositionsConfig()
    ctrl_config    : CtrlConfig = CtrlConfig()
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Method to save back the configuration     
    def cfgdict(self, exclude=set()):
        
        d = super().cfgdict(exclude=exclude)
        for a in ('initialisation', 'positions'):
            if a not in exclude:
                d[a] = getattr(self, a).cfgdict()   
        return d






##### ############
# Sequence
class INITSEQ(int, Enum):
    END = 0
    FIND_INDEX = 1
    FIND_REF_LE = 2
    FIND_REF_UE = 3
    FIND_LHW = 4
    FIND_UHW = 5

    DELAY = 6
    MOVE_ABS = 7
    MOVE_REL = 8
    CALIB_ABS = 9
    CALIB_REL = 10
    CALIB_SWITCH = 11

for E,v1,v2 in [
    ( INITSEQ.END,  "", "" ),
    ( INITSEQ.FIND_INDEX, "Fast Vel", "Slow Vel" ),
    ( INITSEQ.FIND_REF_LE, "Fast Vel", "Slow Vel" ),
    ( INITSEQ.FIND_REF_UE, "Fast Vel", "Slow Vel" ),
    ( INITSEQ.FIND_LHW, "Fast Vel", "Slow Vel" ),
    ( INITSEQ.FIND_UHW, "Fast Vel", "Slow Vel" ),
    ( INITSEQ.DELAY, "Delay [ms]", "" ),
    ( INITSEQ.MOVE_ABS, "Vel", "Pos" ),
    ( INITSEQ.MOVE_REL, "Vel", "Pos" ),
    ( INITSEQ.CALIB_ABS, "Pos", "" ),
    ( INITSEQ.CALIB_REL, "Pos", "" ),
    ( INITSEQ.CALIB_SWITCH, "Pos", "" ),
]:
    setattr(E, "var1", v1)
    setattr(E, "var2", v2)
del E,v1,v2



class STATE(int, Enum):
    IDLE = 20
    RESET_AXIS = 30
    SET_POS = 40
    INIT = 50
    SAVE_TO_NOVRAM = 55
    CLEAR_NOVRAM = 57
    MOVE_ABS = 60
    MOVE_OPTIMISED = 61
    MOVE_VEL = 70
    CTRL_BRAKE = 75
    STOP = 80


class MOTOR_COMMAND(int, Enum):
    NONE = _inc(0)
    INITIALISE = _inc()
    SET_POSITION = _inc()
    MOVE_ABSOLUTE = _inc()
    MOVE_RELATIVE = _inc()
    MOVE_VELOCITY = _inc()
    NEW_VELOCITY = _inc()
    NEW_POSITION = _inc()
    CLEAR_NOVRAM = _inc()

@record_class
class MotorCommand(BaseParser):
    class Config(BaseParser.Config):
        type = "MotorCommand"
    @staticmethod
    def parse(value, config):
        if isinstance(value, str):
            value =  getattr(MOTOR_COMMAND, value)
        return Int32(value)

class DIRECTION(int, Enum):
    POSITIVE = _inc(1)
    SHORTEST = _inc()
    NEGATIVE = _inc()
    CURRENT  = _inc()

@record_class
class Direction(BaseParser):
    class Config(BaseParser.Config):
        type = "Direction"
    @staticmethod
    def parse(value, config):
        if isinstance(value, str):
            value =  getattr(DIRECTION, value)
        return Int16(value)




class AXIS_TYPE(int, Enum):
    LINEAR = 1
    CIRCULAR =2
    CIRCULAR_OPTIMISED = 3



def axis_type(axis_type):
    """ return always a axis_type int number from a number or a string
    
    Raise a ValueError if the input string does not match axis type
    Example:
        axis_type('LINEAR') == 1
        axis_type(1) == 1
    """
    if isinstance(axis_type, str):
        try:
            axis_type = getattr(AXIS_TYPE, axis_type) 
        except AttributeError:
            raise ValueError(f'Unknown AXIS type {axis_type!r}')
    return Int32(axis_type)

# a parser class for axis type
@record_class
class AxisType(BaseParser):
    class Config(BaseParser.Config):
        type: str = "AxisType"
    @staticmethod
    def parse(value, config):
        return axis_type(value)   





# 
#   __                  _   _                 
#  / _|_   _ _ __   ___| |_(_) ___  _ __  ___ 
# | |_| | | | '_ \ / __| __| |/ _ \| '_ \/ __|
# |  _| |_| | | | | (__| |_| | (_) | | | \__ \
# |_|  \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
# 

def init_sequence_to_cfg(initialisation, INITSEQ=INITSEQ):
    """ from a config initialisation dict return a dictionary of key/value for .cfg interface """            
    
    
    # set the init sequence    
    cfg_dict = {} 
    
    init_dict = initialisation.dict(exclude_none=True, exclude_unset=True)
    if not "sequence" in init_dict:        
        return cfg_dict
    
    # reset all sequence variable
    for i in range(1,11):
        cfg_dict["init_seq{}_action".format(i)] = INITSEQ.END.value
        cfg_dict["init_seq{}_value1".format(i)] = 0.0
        cfg_dict["init_seq{}_value2".format(i)] = 0.0
        
    for stepnum, step_name in enumerate(initialisation.sequence, start=1):
        step = getattr(initialisation, step_name)
        cfg_dict["init_seq%d_action"%stepnum] = step.index
        cfg_dict["init_seq%d_value1"%stepnum] = step.value1
        cfg_dict["init_seq%d_value2"%stepnum] = step.value2    
    return cfg_dict


#  ___       _             __                     
# |_ _|_ __ | |_ ___ _ __ / _| __ _  ___ ___  ___ 
#  | || '_ \| __/ _ \ '__| |_ / _` |/ __/ _ \/ __|
#  | || | | | ||  __/ |  |  _| (_| | (_|  __/\__ \
# |___|_| |_|\__\___|_|  |_|  \__,_|\___\___||___/
                                                

NP = VltDevice.Node.prop

@record_class
class StatInterface(VltDevice.StatInterface):
    STATE = STATE
    class Config(VltDevice.StatInterface.Config):
        type = "VltMotorStat"
    state = NP('state', suffix= 'nState' )
    pos_target = NP('pos_target', suffix= 'stat.lrPosTarget' )
    pos_actual = NP('pos_actual', suffix= 'stat.lrPosActual' )
    vel_actual = NP('vel_actual', suffix= 'stat.lrVelActual' )
    vel_target = NP('vel_target', suffix= 'stat.lrVelTarget' )
    axis_status = NP('axis_status', suffix= 'stat.nAxisStatus' )
    backlash_step = NP('backlash_step', suffix= 'stat.nBacklashStep' )
    last_command = NP('last_command', suffix= 'stat.nLastCommand' )
    error_code = NP('error_code', suffix= 'stat.nErrorCode' )
    error_text = NP('error_text', suffix= 'stat.sErrorText' )
    init_step = NP('init_step', suffix= 'stat.nInitStep' )
    init_action = NP('init_action', suffix= 'stat.nInitAction' )
    info_data1 = NP('info_data1', suffix= 'stat.nInfoData1' )
    info_data2 = NP('info_data2', suffix= 'stat.nInfoData2' )
    local = NP('local', suffix= 'stat.bLocal' )
    enabled = NP('enabled', suffix= 'stat.bEnabled' )
    initialised = NP('initialised', suffix= 'stat.bInitialised' )
    ref_switch = NP('ref_switch', suffix= 'stat.bRefSwitch' )
    at_max_position = NP('at_max_position', suffix= 'stat.bAtMaxPosition' )
    at_min_position = NP('at_min_position', suffix= 'stat.bAtMinPosition' )
    limit_switch_positive = NP('limit_switch_positive', suffix= 'stat.bLimitSwitchPositive' )
    limit_switch_negative = NP('limit_switch_negative', suffix= 'stat.bLimitSwitchNegative' )
    brake_active = NP('brake_active', suffix= 'stat.bBrakeActive' )
    max_position = NP('max_position', suffix= 'stat.lrMaxPositionValue' )
    min_position = NP('min_position', suffix= 'stat.lrMinPositionValue' )
    mode = NP('mode', suffix= 'stat.nMode' )
    axis_ready = NP('axis_ready', suffix= 'stat.bAxisReady' )
    moving_abs = NP('moving_abs', suffix= 'stat.bMovingAbs' )
    moving_vel = NP('moving_vel', suffix= 'stat.bMovingVel' )
    changing_vel = NP('changing_vel', suffix= 'stat.bChangingVel' )
    


    @NodeAlias.prop("check", nodes=["error_code", "error_text"])
    def check(self, erc, ert):
        """ This node always return True but raise an error in case of device in error """
        if erc:
            raise RuntimeError(f"Error {erc}: {ert}")
        return True

    @NodeAlias1.prop("state_txt", node="state")
    def state_txt(self, state):
        return self.STATE(state).name 

    @NodeAlias.prop("movement_finished", nodes=["state", "check"])
    def movement_finished(self, state, c):
        return state not in [STATE.MOVE_ABS, STATE.MOVE_OPTIMISED, STATE.MOVE_VEL, STATE.INIT]

    @NodeAlias.prop("initialisation_finished", nodes=["initialised", "check"])
    def initialisation_finished(self, initialised, c):
        return initialised

    @NodeAlias.prop("enable_finished", nodes=["enabled", "check"])
    def enable_finished(self, enabled, c):
        return enabled
    
    _mot_positions = None# will be overwriten by Motor 
    @NodeAlias1.prop("pos_name", node="pos_actual")
    def pos_name(self, pos_actual):
        if not self._mot_positions: return ''
        positions = self._mot_positions
        tol = positions.tolerance
        for pname, pos in positions.positions.items():
            if abs( pos-pos_actual)<tol:
                return pname
        return ''
   
    
    
    not_initialised = NegNode.prop("not_initialised", node="initialised")

    class Data(VltDevice.StatInterface.Data):
        state :NodeVar[int] = 0 
        pos_target : NodeVar[float] = 0.0 
        pos_actual : NodeVar[float] = 0.0 
        vel_actual : NodeVar[float] = 0.0 
        vel_target : NodeVar[float] = 0.0 
        axis_status :NodeVar[int] = 0 
        backlash_step :NodeVar[int] = 0
        last_command : NodeVar[int] = 0
        error_code: NodeVar[int] = 0
        error_text: NodeVar[str] = ""
        init_step: NodeVar[int] = 0
        init_action : NodeVar[int] = 0
        info_data1  : NodeVar[int] = 0
        info_data2  : NodeVar[int] = 0
        local : NodeVar[bool] = False
        enabled  : NodeVar[bool] = False
        initialised  : NodeVar[bool] = False
        ref_switch  : NodeVar[bool] = False
        at_max_position  : NodeVar[bool] = False
        at_min_position  : NodeVar[bool] = False
        limit_switch_positive  : NodeVar[bool] = False
        limit_switch_negative  : NodeVar[bool] = False
        brake_active  : NodeVar[bool] = False
        max_position  : NodeVar[float] = 0.0 
        min_position  : NodeVar[float] = 0.0 
        mode  : NodeVar[int] = 0
        axis_ready  : NodeVar[bool] = False
        moving_abs  : NodeVar[bool] = False
        moving_vel  : NodeVar[bool] = False
        changing_vel  : NodeVar[bool] = False
        

       



@record_class
class CfgInterface(VltDevice.CfgInterface):
    class Config(VltDevice.CfgInterface.Config):
        type = "VltMotorCfg"
    scale_factor = NP('scale_factor', suffix= 'cfg.lrScaleFactor' )
    accel = NP('accel', suffix= 'cfg.lrAccel' )
    decel = NP('decel', suffix= 'cfg.lrDecel' )
    jerk = NP('jerk', suffix= 'cfg.lrJerk' )
    backlash = NP('backlash', suffix= 'cfg.lrBacklash' )
    velocity = NP('velocity', suffix= 'cfg.lrDefaultVelocity' )
    max_pos = NP('max_pos', suffix= 'cfg.lrMaxPosition' )
    min_pos = NP('min_pos', suffix= 'cfg.lrMinPosition' )
    tolerence = NP('tolerence', suffix= 'cfg.lrTolerance' )
    tolerence_enc = NP('tolerence_enc', suffix= 'cfg.lrToleranceEnc' )
    axis_type = NP('axis_type', suffix= 'cfg.nTypeAxis', parser= 'AxisType' )
    tout_init = NP('tout_init', suffix= 'cfg.tTimeoutInit', parser= 'UaInt32' )
    tout_move = NP('tout_move', suffix= 'cfg.tTimeoutMove', parser= 'UaInt32' )
    tout_switch = NP('tout_switch', suffix= 'cfg.tTimeoutSwitch', parser= 'UaInt32' )
    brake = NP('brake', suffix= 'cfg.bUseBrake' )
    low_brake = NP('low_brake', suffix= 'cfg.bActiveLowBrake' )
    active_low_lstop = NP('active_low_lstop', suffix= 'cfg.bArrActiveLow[0].bActiveLow' )
    active_low_lhw = NP('active_low_lhw', suffix= 'cfg.bArrActiveLow[1].bActiveLow' )
    active_low_ref = NP('active_low_ref', suffix= 'cfg.bArrActiveLow[2].bActiveLow' )
    active_low_index = NP('active_low_index', suffix= 'cfg.bArrActiveLow[3].bActiveLow' )
    active_low_uhw = NP('active_low_uhw', suffix= 'cfg.bArrActiveLow[4].bActiveLow' )
    active_low_ustop = NP('active_low_ustop',   suffix= 'cfg.bArrActiveLow[5].bActiveLow' )
    init_seq1_action = NP('init_seq1_action',   suffix= 'cfg.strArrInitSeq[1].nAction', parser= 'UaInt32' )
    init_seq1_value1 = NP('init_seq1_value1',   suffix= 'cfg.strArrInitSeq[1].lrValue1' )
    init_seq1_value2 = NP('init_seq1_value2',   suffix= 'cfg.strArrInitSeq[1].lrValue2' )
    init_seq2_action = NP('init_seq2_action',   suffix= 'cfg.strArrInitSeq[2].nAction', parser= 'UaInt32' )
    init_seq2_value1 = NP('init_seq2_value1',   suffix= 'cfg.strArrInitSeq[2].lrValue1' )
    init_seq2_value2 = NP('init_seq2_value2',   suffix= 'cfg.strArrInitSeq[2].lrValue2' )
    init_seq3_action = NP('init_seq3_action',   suffix= 'cfg.strArrInitSeq[3].nAction', parser= 'UaInt32' )
    init_seq3_value1 = NP('init_seq3_value1',   suffix= 'cfg.strArrInitSeq[3].lrValue1' )
    init_seq3_value2 = NP('init_seq3_value2',   suffix= 'cfg.strArrInitSeq[3].lrValue2' )
    init_seq4_action = NP('init_seq4_action',   suffix= 'cfg.strArrInitSeq[4].nAction', parser= 'UaInt32' )
    init_seq4_value1 = NP('init_seq4_value1',   suffix= 'cfg.strArrInitSeq[4].lrValue1' )
    init_seq4_value2 = NP('init_seq4_value2',   suffix= 'cfg.strArrInitSeq[4].lrValue2' )
    init_seq5_action = NP('init_seq5_action',   suffix= 'cfg.strArrInitSeq[5].nAction', parser= 'UaInt32' )
    init_seq5_value1 = NP('init_seq5_value1',   suffix= 'cfg.strArrInitSeq[5].lrValue1' )
    init_seq5_value2 = NP('init_seq5_value2',   suffix= 'cfg.strArrInitSeq[5].lrValue2' )
    init_seq6_action = NP('init_seq6_action',   suffix= 'cfg.strArrInitSeq[6].nAction', parser= 'UaInt32' )
    init_seq6_value1 = NP('init_seq6_value1',   suffix= 'cfg.strArrInitSeq[6].lrValue1' )
    init_seq6_value2 = NP('init_seq6_value2',   suffix= 'cfg.strArrInitSeq[6].lrValue2' )
    init_seq7_action = NP('init_seq7_action',   suffix= 'cfg.strArrInitSeq[7].nAction', parser= 'UaInt32' )
    init_seq7_value1 = NP('init_seq7_value1',   suffix= 'cfg.strArrInitSeq[7].lrValue1' )
    init_seq7_value2 = NP('init_seq7_value2',   suffix= 'cfg.strArrInitSeq[7].lrValue2' )
    init_seq8_action = NP('init_seq8_action',   suffix= 'cfg.strArrInitSeq[8].nAction', parser= 'UaInt32' )
    init_seq8_value1 = NP('init_seq8_value1',   suffix= 'cfg.strArrInitSeq[8].lrValue1' )
    init_seq8_value2 = NP('init_seq8_value2',   suffix= 'cfg.strArrInitSeq[8].lrValue2' )
    init_seq9_action = NP('init_seq9_action',   suffix= 'cfg.strArrInitSeq[9].nAction', parser= 'UaInt32' )
    init_seq9_value1 = NP('init_seq9_value1',   suffix= 'cfg.strArrInitSeq[9].lrValue1' )
    init_seq9_value2 = NP('init_seq9_value2',   suffix= 'cfg.strArrInitSeq[9].lrValue2' )
    init_seq10_action = NP('init_seq10_action', suffix= 'cfg.strArrInitSeq[10].nAction', parser= 'UaInt32' )
    init_seq10_value1 = NP('init_seq10_value1', suffix= 'cfg.strArrInitSeq[10].lrValue1' )
    init_seq10_value2 = NP('init_seq10_value2', suffix= 'cfg.strArrInitSeq[10].lrValue2' )


    class Data(VltDevice.CfgInterface.Data):

        scale_factor: NodeVar[float] = 1.0
        accel:        NodeVar[float] = 30.0
        decel :       NodeVar[float] = 30.0
        jerk :        NodeVar[float] = 100.0
        brake:            NodeVar[bool] = False    
        backlash:         NodeVar[float] = 0.0
        axis_type:        NodeVar[int] = 0 
        
        velocity:         NodeVar[float] = 0.0
        max_pos:          NodeVar[float] = 0.0      
        min_pos:          NodeVar[float] = 0.0      
        
        tolerence: NodeVar[float] = 1.0
        tolerence_enc: NodeVar[int] = 100
        
        tout_init:          NodeVar[int] = 0 
        tout_move:          NodeVar[int] = 0 
        tout_switch:        NodeVar[int] = 0 


        low_brake:          NodeVar[bool] = False 
        active_low_lstop:   NodeVar[bool] = False 
        active_low_lhw:     NodeVar[bool] = False 
        active_low_ref:     NodeVar[bool] = False 
        active_low_index:   NodeVar[bool] = False 
        active_low_uhw:     NodeVar[bool] = False 
        active_low_ustop:   NodeVar[bool] = False 

   
        init_seq1_action:  NodeVar[int] = 0
        init_seq1_value1:  NodeVar[float] = 0.0
        init_seq1_value2:  NodeVar[float] = 0.0
        
        init_seq2_action:  NodeVar[int] = 0
        init_seq2_value1:  NodeVar[float] = 0.0
        init_seq2_value2:  NodeVar[float] = 0.0
        
        init_seq3_action:  NodeVar[int] = 0
        init_seq3_value1:  NodeVar[float] = 0.0
        init_seq3_value2:  NodeVar[float] = 0.0
        
        init_seq4_action:  NodeVar[int] = 0
        init_seq4_value1:  NodeVar[float] = 0.0
        init_seq4_value2:  NodeVar[float] = 0.0
        
        init_seq5_action:  NodeVar[int] = 0
        init_seq5_value1:  NodeVar[float] = 0.0
        init_seq5_value2:  NodeVar[float] = 0.0

        init_seq6_action:  NodeVar[int] = 0
        init_seq6_value1:  NodeVar[float] = 0.0
        init_seq6_value2:  NodeVar[float] = 0.0

        init_seq7_action:  NodeVar[int] = 0
        init_seq7_value1:  NodeVar[float] = 0.0
        init_seq7_value2:  NodeVar[float] = 0.0

        init_seq8_action:  NodeVar[int] = 0
        init_seq8_value1:  NodeVar[float] = 0.0
        init_seq8_value2:  NodeVar[float] = 0.0

        init_seq9_action:  NodeVar[int] = 0
        init_seq9_value1:  NodeVar[float] = 0.0
        init_seq9_value2:  NodeVar[float] = 0.0

        init_seq10_action:  NodeVar[int] = 0
        init_seq10_value1:  NodeVar[float] = 0.0
        init_seq10_value2:  NodeVar[float] = 0.0    
        

@record_class
class CtrlInterface(VltDevice.CtrlInterface):
    class Config(VltDevice.CtrlInterface.Config):
        type = "VltMotorCtrl"
    command =   NP('command',   suffix= 'ctrl.nCommand',    parser= MotorCommand )
    direction = NP('direction', suffix= 'ctrl.nDirection',  parser= Direction )
    position = NP('position',   suffix= 'ctrl.lrPosition',  parser= 'UaDouble' )
    velocity = NP('velocity',   suffix= 'ctrl.lrVelocity',  parser= 'UaDouble' )
    stop =     NP('stop',       suffix= 'ctrl.bStop',       parser=  bool  )
    reset =    NP('reset',      suffix= 'ctrl.bResetError', parser=  bool  )
    disable =  NP('disable',    suffix= 'ctrl.bDisable',    parser=  bool  )
    enable =   NP('enable',    suffix= 'ctrl.bEnable',      parser=  bool  )
    execute =  NP('execute',   suffix= 'ctrl.bExecute',     parser=  bool  )


# __     ___ _   __  __       _             
# \ \   / / | |_|  \/  | ___ | |_ ___  _ __ 
#  \ \ / /| | __| |\/| |/ _ \| __/ _ \| '__|
#   \ V / | | |_| |  | | (_) | || (_) | |   
#    \_/  |_|\__|_|  |_|\___/ \__\___/|_|   
                                          

@record_class
class VltMotor(VltDevice):
    
    Config = VltMotorConfig

    CfgInterface = CfgInterface
    StatInterface = StatInterface
    CtrlInterface = CtrlInterface 
    
    COMMAND = MOTOR_COMMAND
    DIRECTION = DIRECTION
    INITSEQ = INITSEQ
    
    class Data(VltDevice.Data):
        StatData = StatInterface.Data
        CfgData = CfgInterface.Data

        stat: StatData = StatData()
        cfg: CfgData = CfgData()

    cfg  = CfgInterface.prop('cfg')
    ctrl = CtrlInterface.prop('ctrl')

    @StatInterface.prop('stat')    
    def stat(self, interface):
        # need to add the motor_position to the stat interface instance 
        interface._mot_positions = self.config.positions
        
    


    def get_configuration(self, exclude_unset=True,  **kwargs) -> Dict[VltDevice.Node,Any]:
        """  return a node/value pair dictionary ready to be uploaded 
        
        The node/value dictionary represent the device configuration. 
        
        Args:
            **kwargs : name/value pairs pointing to cfg.name node
                      This allow to change configuration on the fly
                      without changing the config file. 
        """
        
        config = self._config 
        
        ctrl_config = config.ctrl_config
        # just update what is in ctrl_config, this should work for motor 
        # one may need to check parse some variable more carefully       
        values = ctrl_config.dict(exclude_none=True, exclude_unset=exclude_unset)
        cfg_dict = {self.cfg.get_node(k):v for k,v in  values.items() }
        cfg_dict[self.ignored] = self.config.ignored 
        cfg_dict.update({self.cfg.get_node(k):v for k,v in  kwargs.items() })
        
        init_cfg = init_sequence_to_cfg(config.initialisation, self.INITSEQ)
        cfg_dict.update({self.cfg.get_node(k):v for k,v in init_cfg.items()})
        
        # transform axis type to number 
        if self.cfg.axis_type in cfg_dict:
            axis_type = cfg_dict[self.cfg.axis_type] 
            cfg_dict[self.cfg.axis_type] =  getattr(AXIS_TYPE, axis_type) if isinstance(axis_type, str) else axis_type
        ###
        # Set the new config value to the device 
        return cfg_dict
    
    @property
    def velocity(self) -> float:
        return self.config.ctrl_config.velocity
   

    def check(self):
        """Check if the device is in error. Raise an error in case it is True """
        return self.stat.check.get()

    def stop(self):
        self.ctrl.stop.set(True)
    
    def reset(self):
        self.ctrl.reset.set(True)
        return self.stat.not_initialised
    
    def enable(self):
        self.ctrl.enable.set(True)
        return self.stat.enable_finished
    
    def disable(self):
        self.ctrl.disable.set(True)
        return self.stat.enabled


    def init(self):
        upload({ 
            self.ctrl.execute : True, 
            self.ctrl.command : self.COMMAND.INITIALISE
            })
        return self.stat.initialisation_finished

    def move_abs(self, pos, vel=None):
        
        vel = self.velocity if vel is None else vel
        upload({
            self.ctrl.execute : True, 
            self.ctrl.position: pos, 
            self.ctrl.velocity: vel, 
            self.ctrl.command : self.COMMAND.MOVE_ABSOLUTE
            })
        return self.stat.movement_finished

    def move_rel(self, pos, vel=None):

        vel = self.velocity if vel is None else vel
        upload({
            self.ctrl.execute : True, 
            self.ctrl.position: pos, 
            self.ctrl.velocity: vel, 
            self.ctrl.command : self.COMMAND.MOVE_RELATIVE
            })
        return self.stat.movement_finished  

    def move_vel(self, vel):
        direction = self.DIRECTION.POSITIVE if vel>0 else self.DIRECTION.NEGATIVE
        vel = abs(vel)
        upload({
            self.ctrl.execute : True, 
            self.ctrl.velocity: vel, 
            self.ctrl.direction: direction, 
            self.ctrl.command: self.COMMAND.MOVE_VELOCITY
            })
        return None
    
    def get_pos_target_of_name(self, name: str) -> float:
        """return the configured target position of a given pos name or raise error"""
        try:
            position = getattr(self.config.positions, name)
        except AttributeError:
            raise ValueError('unknown posname %r'%name)
        return position

    def get_name_of_pos(self, pos_actual: float) -> str:
        """ Retrun the name of a position from a position as input or ''
        
        Example:
            m.get_name_of( m.stat.pos_actual.get() )
        """
        positions = self.config.positions    
        tol = positions.tolerance
        
        for pname, pos in positions.positions.items():
            if abs( pos-pos_actual)<tol:
                return pname
        return ''

    def move_name(self, name, vel=None) -> VltDevice.Node:
        """ move motor to a named position 
        
        Args:
           name (str): named position
           vel (float):   target velocity for the movement
        """
        absPos = self.get_pos_target_of_name(name)
        return self.move_abs(absPos, vel)
        

    def set_pos(self, pos):
        """ Set the curent position value """
        upload({
            self.ctrl.execute : True, 
            self.ctrl.position: pos, 
            self.ctrl.command : self.COMMAND.SET_POSITION
            })





from PyQt5 import uic
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton

from .io import find_ui
from pydevmgr_core import NodeVar
from .elt_device_ctrl import EltDeviceCtrl
from pydevmgr_core_qt import record_widget_factory, BaseUiLinker
from pydevmgr_core import DequeNode, UnixTimeNode
from pydantic import Field
import pyqtgraph as pg
import numpy as np 
from enum import IntEnum

#####
# Define all the data used at fron-end 
# The NodeVar annotation indicate that the data will be taken from a device node see `pydevmgr.DataLink`


# create a node returning the unix  time 
time = UnixTimeNode('time')


plot_nodes = [time, "pos_actual", "pos_error", "vel_actual"]

class MotorPlotStatData(EltDeviceCtrl.Data.StatData):
    # DeviceStatData contain state, substate, error code and txt information 
    # Add here all NodeVar needed for the motor GUI   
    
    pos_actual: NodeVar[float] = 0.0
    pos_error: NodeVar[float]  = 0.0
    plot_data: NodeVar[list] = Field([], node=DequeNode.prop('plot_data' , nodes=plot_nodes, trigger_index=-1))


###
# Example of use:
#  data = MotorCtrlData()
#  dl = DataLink(motor, data)
#  dl.update() # update data from server(s)
#  print(data.stat.pos_actual, data.stat.pos_error) 
class MotorPlotData(EltDeviceCtrl.Data):
    StatData = MotorPlotStatData
    stat: StatData = StatData()


# We create a widget holding the two plots 
class MotorPlotWidget(QWidget):
    def __init__(self,*args, **kwargs):
        super().__init__(*args, **kwargs)
                
        layout = QVBoxLayout()
        self.setLayout(layout)
        
        # plots
        g_layout = QHBoxLayout()
        plots = {
          'pos' :       pg.PlotWidget(title='Position'), 
          'pos_error' : pg.PlotWidget(title='Error'), 
        }
        g_layout.addWidget(plots['pos'])
        g_layout.addWidget(plots['pos_error'])        
        layout.addLayout(g_layout)
        
        # a Clear button 
        clear_button = QPushButton()
        clear_button.setText('Clear')
        layout.addWidget(clear_button)
        
        self.clear_button = clear_button
        self.plots = plots





    
class MotorPlot(BaseUiLinker):
    # A widget constructor for default widget 
    Widget = MotorPlotWidget
    # A data constructor 
    Data = MotorPlotData
    
    # Basicaly 3 methods needs to be implemented 
    # - .init_vars() :  build the variable handler function to their type and widget
    # - .update(data) :   update widget with new data 
    # - .connect_device(device, data) : this is where commands has to be connected to widget buttons etc...
    start_time = None 
    def init_vars(self):
        super().init_vars()
        self.outputs.plots = self.widget.plots

    def update(self, data: MotorPlotData) -> None:
        """ Update the widget (the plots in our case) """
        plotdata = np.asarray(data.stat.plot_data)
        
        if len(plotdata):
            if self.start_time is None:
                self.start_time = min(plotdata[:,0])
                
            p = self.outputs.plots['pos']
            p.clear()        
            p.plot(plotdata[:,0]- self.start_time, plotdata[:,1])
            
            p = self.outputs.plots['pos_error']
            p.clear()        
            p.plot(plotdata[:,0]- self.start_time, plotdata[:,2])
 
    
      
    def connect_device(self, device, data):

        def clear_plotdata():
            self.start_time = None
            data.stat.plotdata.clear()
            self.outputs.plots['pos'].clear()
            self.outputs.plots['pos_error'].clear()
        
         # we could also do self.widget.clear_button.clicked.connect(clear_plotdata)
         # but the actions object allow to collect all widget connection and unlink them properly 
         # when the window is closed for instance
        self.actions.add(clear_plotdata).connect_button(self.widget.clear_button)           


record_widget_factory("plot", "Motor", MotorPlot)





from pydevmgr_vlt.vltmotor import StatInterface
from .vltdevice import VltDevice
from pydevmgr_core import BaseParser, record_class, upload, NodeAlias, NodeVar
from pydevmgr_ua import Int32
from enum import Enum 
from typing import Union, List, Dict


N_AI, N_DI, N_NI, N_TI  = [8]*4 
N_AO, N_DO, N_NO, N_TO  = [8]*4

class STATUS(int, Enum):
    OK = 0
    ERROR = 1



class COMMAND(int, Enum):
    NONE = 0
    INITIALISE = 1
    ACTIVATE = 2

@record_class
class IoDevCommand(BaseParser):
    class Config(BaseParser.Config):
        type = "IoDevCommand"
    @staticmethod
    def parse(value, config):
        if isinstance(value, str):
            value =  getattr(COMMAND, value)
        return Int32(value)

NP = VltDevice.Node.prop





class VltIoDevStatInterface(VltDevice.StatInterface):
    class Config(VltDevice.StatInterface.Config):
        type = "VltIoDevStatInterface"
    
    initialised = NP('initialised', suffix= 'stat.bInitialised' )
    last_command = NP('last_command', suffix= 'stat.nLastCommand' )
    error_code = NP('error_code', suffix= 'stat.nErrorCode' )
    error_text = NP('error_text', suffix= 'stat.sErrorText' )
    status = NP('status', suffix= 'stat.nStatus' )   
    # Other Nodes arre added bellow 
    
    @NodeAlias.prop( 'di_all', [f'di_{i}' for i in range(N_DI) ] )
    def di_all(self, *flags):
        return flags
    @NodeAlias.prop( 'ai_all', [f'ai_{i}' for i in range(N_AI)] )
    def ai_all(self, *values):
        return values
    @NodeAlias.prop( 'ni_all', [f'ni_{i}' for i in range(N_NI)] )
    def ni_all(self, *values):
        return values
    @NodeAlias.prop( 'ti_all', [f'ti_{i}' for i in range(N_TI)] )
    def ti_all(self, *values):
        return values

    class Data(VltDevice.StatInterface.Data):
        initialised : NodeVar[bool] = False
        last_command: NodeVar[COMMAND] = COMMAND.NONE
        error_code: NodeVar[int] = 0 
        error_text: NodeVar[str] = ""
        status: NodeVar[STATUS] = STATUS.OK
        di_all: NodeVar[list] = []
        ai_all: NodeVar[list] = []
        ni_all: NodeVar[list] = []
        ti_all: NodeVar[list] = []


# ADD all the DI, NI, etc nodes 
for i in range(N_DI):
    setattr( VltIoDevStatInterface, f'di_{i}', NP(f'di_{i}', suffix= f'stat.arr_DI[{i}].bValue') )
for i in range(N_AI):
    setattr( VltIoDevStatInterface, f'ai_{i}', NP(f'ai_{i}', suffix= f'stat.arr_AI[{i}].lrValue') )
for i in range(N_NI):
    setattr( VltIoDevStatInterface, f'ni_{i}', NP(f'ni_{i}', suffix= f'stat.arr_NI[{i}].nValue') )
for i in range(N_TI):
    setattr( VltIoDevStatInterface, f'ti_{i}', NP(f'ti_{i}', suffix= f'stat.arr_TI[{i}].sValue') )

    

class VltIoDevCtrlInterface(VltDevice.CtrlInterface):
    class Config(VltDevice.CtrlInterface.Config):
        type = "VltIoDevCtrlInterface"

    execute = NP('execute', suffix= 'ctrl.bExecute', parser= 'bool' )
    command = NP('command', suffix= 'ctrl.nCommand', parser= 'IoDevCommand' )
    

# ADD all the DI, NI, etc nodes 
for i in range(N_DO):
    setattr( VltIoDevCtrlInterface, f'do_{i}', NP(f'do_{i}', suffix= f'ctrl.arr_DO[{i}].bValue') )
for i in range(N_AO):
    setattr( VltIoDevCtrlInterface, f'ao_{i}', NP(f'ao_{i}', suffix= f'ctrl.arr_AO[{i}].lrValue') )
for i in range(N_NO):
    setattr( VltIoDevCtrlInterface, f'no_{i}', NP(f'no_{i}', suffix= f'ctrl.arr_NO[{i}].nValue') )
for i in range(N_TO):
    setattr( VltIoDevCtrlInterface, f'to_{i}', NP(f'to_{i}', suffix= f'ctrl.arr_TO[{i}].sValue') )

del NP


# __     ___ _   ___      ____             
# \ \   / / | |_|_ _|___ |  _ \  _____   __
#  \ \ / /| | __|| |/ _ \| | | |/ _ \ \ / /
#   \ V / | | |_ | | (_) | |_| |  __/\ V / 
#    \_/  |_|\__|___\___/|____/ \___| \_/  
                                         


@record_class
class VltIoDev(VltDevice):
    COMMAND = COMMAND
    StatInterface =  VltIoDevStatInterface
    CtrlInterface = VltIoDevCtrlInterface

    stat = StatInterface.prop("stat")
    ctrl = CtrlInterface.prop("ctrl")
    
    class Config(VltDevice.Config):
        type = "VltIoDev"
    
    class Data(VltDevice.Data):
        StatData = StatInterface.Data
        stat: StatData = StatData()
         

    def set_do(self, flags: Union[List[bool], Dict[int,bool]]):
        """ set digital output flags 
        
        Args:
            flags (list, or dict): list of bool or a dictionary of digital output index (starting from 0) and flag pair
             
        Exemple::

            io.set_do( [False]*8 ) # set all to zero 
            io.set_do( [True, False] ) # set do_0 and do_1 to True and False respectively (others are unchaged)
            io.set_do( {3:True, 4:True} ) # set do_4 and do_4 to True (others are unchanged)
        """
        if not isinstance(flags, dict):
            it = enumerate(flags)
        else:
            it = flags.items()

        ctrl = self.ctrl 
        n_f = { ctrl.get_node( "do_{}".format(i) ):f for i,f in it }
        
        n_f.update( {ctrl.execute:True, ctrl.command :self.COMMAND.ACTIVATE} )
        upload(n_f)
    
    def set_ao(self, values: Union[List[bool], Dict[int,bool]]):
        """ set degital output values 
        
        Args:
            flags (list, or dict): list of float or a dictionary of analog output index (starting from 0) and value  pair
             
        Exemple::

            io.set_ao( [0.0]*8 ) # set all to zero 
            io.set_ao( [32, 64] ) # set do_0 and do_1 to 32 and 64 respectively (others are unchaged)
            io.set_ao( {3:128, 4:128} ) # set do_4 and do_4 to 128 (others are unchanged)
        """

        if not isinstance(values, dict):
            it = enumerate(values)
        else:
            it = values.items()

        ctrl = self.ctrl 
        n_f = { ctrl.get_node( "do_{}".format(i) ):f for i,f in it }
        
        n_f.update( {ctrl.execute:True, ctrl.command :self.COMMAND.ACTIVATE} )
        upload(n_f)
    



        

        
            





# Include the core objects in this package 
from pydevmgr_core import *
from pydevmgr_ua import  (Int16, Int32, Int64, UInt16, UInt32, UInt64, Float, Double,
                          INT  , DINT , LINT , UINT  , UDINT , ULINT  , REAL, LREAL)


from .vltdevice import VltDevice
from .vltmotor import VltMotor
from .vltiodev import VltIoDev





from pydantic import BaseModel
from PyQt5.QtWidgets import QFrame
from pydevmgr_core_qt import BaseUiLinker, record_widget_factory

        
        
class EltDeviceCfgUi(QFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # What else ? 
    
        
class EltDeviceData(BaseModel):
    name: str = "" # device name  
            
# place holder for some base setup for config pannels 
# not sure we have any             
class EltDeviceCfg(BaseUiLinker):
    Widget = EltDeviceCfgUi
    Data = EltDeviceData
            
record_widget_factory("cfg", "Elt", EltDeviceCfg) 




from PyQt5 import uic
from PyQt5.QtWidgets import QFrame
from .io import find_ui
from pydevmgr_elt import Motor, DataLink

from pydevmgr_core_qt import record_widget_factory, InputOutputs, get_style, STYLE
from .elt_device_cfg import EltDeviceCfg
import math 

class MotorCfgData(EltDeviceCfg.Data):
    CfgData = Motor.Data.CfgData
    
    cfg: CfgData = CfgData()     



def link_init_action():
    pass
        
def float_nan(t):
    try:
        return float(t)
    except (ValueError, TypeError):
        return None

def int_nan(t):
    try:
        return int(t)
    except (ValueError, TypeError):
        return None    

def b2t(b):
    return "[X]" if b else "[_]"

def switch_style(w,b):
    style = STYLE.SIMILAR if b else STYLE.DIFFERENT
    w.setStyleSheet(get_style(style))


class InitSeq(InputOutputs.Base):
    def __init__(self, action_enum,  input_action, input_value1, input_value2, output):
        def set_input(action, value1, value2):
            input_action.setCurrentIndex(action)
            input_value1.setText(f"{value1}")
            input_value2.setText(f"{value2}")  
            
        def get():

            try:
                v1 = float(input_value1.text())
            except (TypeError, ValueError):
                v1 = math.nan
            
            try:
                v2 = float(input_value1.text())
            except (TypeError, ValueError):
                v2 = math.nan
 

            a = input_action.currentIndex()          
            return (a, v1, v2)
        
        def set(action, value1, value2):
            action_txt = action_enum(action).name            
            output.setText(f"{action_txt:>13} {value1:.3f} {value2:.3f}")
            
            switch_style(output, self.get() == (action, value1, value2))        
        super().__init__(set_input, set, get)   
            

class MotorCfgUi(QFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        uic.loadUi(find_ui('motor_cfg_frame.ui'), self)    
        
            
class MotorCfg(EltDeviceCfg):
    Widget = MotorCfgUi
    Data = MotorCfgData
        
    first_update = True
    
    def feedback(self, er, msg):
        # a feedback would be nice 
        pass 
    
    def init_vars(self):
        ws = self.widget
        get = lambda a: getattr(self.widget,a)
        
        self.outputs.name = self.outputs.Str( ws.name )
        
        self.io.keys = {}
        for k in ["brake", "low_brake", "check_inpos", "low_inpos",
            "active_low_lstop", "active_low_lhw", "active_low_ref", "active_low_index",
            "active_low_uhw", "active_low_ustop", "exec_pre_init", "exec_post_init",
            "exec_pre_move", "exec_post_move", "disable",  "lock"]: 
            
            self.io.keys[k]  = self.io.Bool(get("in_"+k),get(k))
        
        for k in ["min_pos", "max_pos", "lock_pos", "lock_tolerance", "backlash"]:
            self.io.keys[k]  = self.io.Float(get("in_"+k), get(k))
        
        for k in ["tout_init", "tout_move", "tout_switch"]:
            self.io.keys[k]  = self.io.Int(get("in_"+k), get(k))
            
        
        self.io.keys['axis_type'] = self.io.Enum(Motor.AXIS_TYPE, self.widget.in_axis_type, self.widget.axis_type)
        
        self.io.sequences = {}
        names = [a.name for a in Motor.INITSEQ]
        for i in range(1,11):
            wa = get(f"init{i}_action")
            wa.addItems(names)
            self.io.sequences[i] = InitSeq(Motor.INITSEQ, 
                                           wa,
                                           get(f"init{i}_value1"),
                                           get(f"init{i}_value2"),
                                           get(f"init{i}_actual"), 
                                           )
                                                           
    def update(self, data) -> None:
        """ update the gui from the attached structure data """
        super().update(data)
        
        cfg = data.cfg
        
        if self.first_update:
            self.edit_all(data.cfg)            
            self.first_update = False
        
        for k, v in self.io.keys.items():
            v.set( getattr(cfg, k) )
        
        for i,s in self.io.sequences.items():
            s.set( getattr(cfg, f"init_seq{i}_action"), 
                                      getattr(cfg, f"init_seq{i}_value1"), 
                                      getattr(cfg, f"init_seq{i}_value2"), 
                                      )
        
    
    def setup_ui(self, motor, data):
        super().setup_ui(motor, data)
        
        self.outputs.name.set( data.name or motor.key )
        
        configure = self.actions.add( lambda : self.configure_motor(motor))
        configure.connect_button(self.widget.in_upload_to)
        
        
        def upload_data():
            data = self.new_data()
            
            #download(DataLink(motor.cfg, cfg))
            DataLink(motor, data).download()
            self.edit_all(data.cfg)
            self.update(data)
        upload_data = self.actions.add(upload_data)
        upload_data.connect_button(self.widget.in_download_from)
        
        
        def upload_data_from_config():
            data = self.new_data()
            cfg = data.cfg
            
            dl = DataLink(motor.cfg, cfg)
            nodes = motor.get_configuration()
                        
            dl.download_from_nodes(nodes)                        
            self.edit_all(cfg)
            self.update(data)
        upload_data_from_config = self.actions.add(upload_data_from_config)
        upload_data_from_config.connect_button(self.widget.in_from_config_file)
                 
        self.first_update = True

            
    #####################################################
    def edit_all(self, cfg):
        for k, v in self.io.keys.items():
            v.set_input( getattr(cfg, k) )
        
        for i,s in self.io.sequences.items():
            s.set_input( getattr(cfg, f"init_seq{i}_action"), 
                         getattr(cfg, f"init_seq{i}_value1"), 
                         getattr(cfg, f"init_seq{i}_value2") )
        
        self.widget.repaint()
        
     
    def configure_motor(self, motor: Motor) -> None:
        """ Configure the motor to what is eddited on the GUI """
        data = self.new_data()
        cfg = data.cfg
        dl = DataLink(motor.cfg, cfg)
        dl.download() # load all server values first 
        # Then update the data to what is defined in the GUI
        
        for k, v in self.io.keys.items():
            setattr(cfg, k, v.get())
        
        for i,s in self.io.sequences.items():
            a,v1,v2 = s.get()
            setattr(cfg, f"init_seq{i}_action", a )
            setattr(cfg, f"init_seq{i}_value1", v1)
            setattr(cfg, f"init_seq{i}_value2", v2)
                
        dl.upload()                
        
record_widget_factory("cfg", "Motor", MotorCfg)        
    
        
        
        
            
        
      



from typing import Optional
from pydevmgr_core import NodeAlias1

_enum = -1 
def _inc(i: Optional[int] = None) -> int:
    """ number increment to use in frontend 
    
    _inc(0) # reset increment to 0 and return 0 
    _inc()  # increment and return incremented number 
    """
    global _enum
    _enum = _enum+1 if i is None else i
    return _enum


class NegNode(NodeAlias1):
    def fget(self, value):
        return not value 



from pydantic import BaseModel
from pydevmgr_elt import NodeVar, EltDevice

from PyQt5 import uic
from PyQt5.QtWidgets import QFrame
from pydevmgr_core_qt import BaseUiLinker, record_widget_factory
from .io import find_ui

class EltDeviceStatData(BaseModel):
    substate: NodeVar[int] = 0
    substate_txt: NodeVar[str] = ""
    substate_group: NodeVar[int] = "" 

class EltDeviceData(BaseModel):
    StatData = EltDeviceStatData
    
    stat: StatData = StatData()
    ignored: NodeVar[bool] = False
    name: str = "" # device name  

class EltDeviceLineUi(QFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        uic.loadUi(find_ui('device_line_frame.ui'), self) 

class EltDeviceLine(BaseUiLinker):  
    Widget = EltDeviceLineUi
    Data = EltDeviceData
    def init_vars(self):        
        # these shall get stat data        
        self.outputs.substate = self.outputs.Code(self.widget.substate)
        self.outputs.name = self.outputs.Str(self.widget.name)
        
        self.inputs.ignored = self.inputs.NBool(self.widget.check)
        
    def feedback(self, er, mgs: str=''):     
           
        if er:
            self.widget.state_action.setItemText(0, "!!ERROR!!")
        else:
            self.widget.state_action.setItemText(0, "")    
        
    def update(self, data: EltDeviceData):        
        
        stat = data.stat 
        self.outputs.substate.set( (stat.substate, stat.substate_txt, stat.substate_group) )
        
        if self.inputs.ignored.get() != data.ignored:
            self.inputs.ignored.set_input(data.ignored)
    
    def setup_ui(self, device: EltDevice, data: EltDeviceData): 
        super().setup_ui(device, data)
        
        # The device name is updated here only 
        self.outputs.name.set( data.name or device.key )
        
        # setup the dorpdown menu list for action     
        wa = self.widget.state_action 
        wa.clear() 
        wa.addItem("") # First item is empty and is always activated after an action 
        action_list = [
         ("INIT",   device.init,   []), 
         ("ENABLE", device.enable, []),
         ("DISABLE",device.disable,[]),
         ("RESET",  device.reset,  []) 
        ]
        # After an command put back the menu to the empty first index
        reset = lambda: wa.setCurrentIndex(0) 
                        
        for i, (name,func,inputs) in  enumerate(action_list, start=wa.count()):
            wa.addItem(name)
            action = self.actions.add(func, inputs, after=reset, feedback=self.feedback)
            action.connect_item(wa, i)
        
                
        w = self.widget.check
        w.setChecked(not device.ignored.get())        
                
        self.actions.add(device.ignored.set, [self.inputs.ignored.get], feedback=self.feedback).connect_checkbox(w)
        
record_widget_factory("line", "Elt", EltDeviceLine)  




from pydevmgr_ua import UaDevice 
from pydevmgr_core import record_class, upload, LocalNode
from .vltinterface import VltInterface


@record_class
class VltDevice(UaDevice):
    Node = VltInterface.Node
    Interface = VltInterface
    
    class Config(UaDevice.Config):
        type = "VLT"
        ignored : bool = False

    class StatInterface(VltInterface):
        pass

    class CfgInterface(VltInterface):
        pass
    
    class CtrlInterface(VltInterface):
        pass
    
    stat = StatInterface.prop("stat")
    cfg =  CfgInterface.prop("cfg")
    ctrl = CtrlInterface.prop("ctrl")


    ignored = LocalNode.prop('ignored', default=False)
    

    def get_configuration(self, exclude_unset=True, **kwargs):
        """ return a node/value pair dictionary ready to be uploaded 
        
        The node/value dictionary represent the device configuration. 
        This is directly use by :func:`Device.configure` method. 
        
        This is a generic configuration dictionary and may not work on all devices. 
        This method need to be updated for special devices for instance.   
        
        Args:
            exclude_unset (optional, bool): Default is True. If True value that was left unset in 
                the config will not be included in the configuration
            \**kwargs : name/value pairs pointing to cfg.name node
                      This allow to change configuration on the fly
                      without changing the config file.             
        
        ::
        
            >>> upload( {**motor1.get_configuration(), **motor2.get_configuration()} ) 
        """
        # get values from the ctrl_config Config Model
        # do not include the default values, if they were unset, the PLC will use the default ones
        values = self.config.ctrl_config.dict(exclude_none=True, exclude_unset=exclude_unset)
        cfg_dict = {self.cfg.get_node(k):v for k,v in values.items()}
        cfg_dict[self.ignored] = self.config.ignored 
        cfg_dict.update({self.cfg.get_node(k):v for k,v in kwargs.items()})
        return cfg_dict

    def configure(self, exclude_unset=True, **kwargs):
        """ Configure the whole device in the PLC according to what is defined in the config dictionary 
        
        Quick changes on configuration value can be done by keywords where each key must point to a 
        .cfg.name node. Note that the configuration (as written in file) is always used first before being 
        overwritten by \**kwargs. In other word kwargs are not changing the default configuration  
        
        Args:
            exclude_unset (optional, bool): Default is True. If True value that was left unset in 
                the config will not be included in the configuration

            \**kwargs :  name/value pairs pointing to cfg.name node
                        This allow to quickly change configuration on the fly
                        without changing the config file.
                          
        
        what it does is just:
        
        ::
        
           >>> upload( self.get_condifuration() ) 
        """
        # by default just copy the "ctrl_config" into cfg. This may not work for
        # all devices and should be customized  
        upload(self.get_configuration(exclude_unset=exclude_unset, **kwargs))



import os
import pkg_resources
from pydevmgr_core.io import find_config

pkg_name = 'pydevmgr_vlt_qt'

def find_ui(resource):
    try:
        return find_config(resource)
    except ValueError:
        pass 
    
    if not pkg_resources.resource_exists(pkg_name, os.path.join('uis',resource)):
        raise IOError('coud not find ui file %r'%(resource))        
    return pkg_resources.resource_filename(pkg_name, os.path.join('uis',resource))   





from PyQt5 import uic
from PyQt5.QtWidgets import QFrame
from .io import find_ui
from pydevmgr_core import NodeVar
from .elt_device_line import  EltDeviceLine
from pydevmgr_core_qt import record_widget_factory


class MotorLineStatData(EltDeviceLine.Data.StatData):
    # DeviceStatData contain state, substate, error code and txt information 
    # Add here all NodeVar needed for the motor GUI    
    pos_actual: NodeVar[float] = 0.0
    pos_name: NodeVar[str]     = ""

class MotorLineData(EltDeviceLine.Data):
    StatData = MotorLineStatData
    stat: StatData = StatData()

class MotorLineUi(QFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        uic.loadUi(find_ui('motor_line_frame.ui'), self)

class MotorLine(EltDeviceLine):
    # define the data model (data constructor) for stat     
    Data = MotorLineData
    Widget = MotorLineUi
    
    def init_vars(self):
        super().init_vars()
        
        self.outputs.pos_actual       = self.outputs.Str(self.widget.pos_actual)
        self.inputs.velocity   = self.inputs.Float(self.widget.input_velocity, default=1.0)
        self.inputs.pos_target = self.inputs.Float(self.widget.input_pos_target)
        
    def update(self, data):
        super().update(data)
        
        #self.input_pos_target.update()
        #self.input_velocity.update() 
        stat = data.stat
        
        pos = stat.pos_name if stat.pos_name else "{:.3f}".format(stat.pos_actual)
            
        self.outputs.pos_actual.set( pos )
    
            
    def setup_ui(self, motor, data):
        super().setup_ui(motor, data)
        
        # init some input fields 
        if motor.is_connected():
            self.inputs.velocity.set_input(motor.cfg.velocity.get())
        
        # add more functionality to the dropdown menu, the standards commands are defined in 
        # super().connect_device
        # Here we are adding STOP MOVE ABS MOVE REL and MOVE VEL     
        wa = self.widget.state_action
        # After an command put back the menu to the empty first index
        reset = lambda : wa.setCurrentIndex(0)
        
        wa.addItem("STOP")
        self.actions.add( motor.stop,
                         [], 
                         after=reset, feedback=self.feedback,
                         ).connect_item(wa, wa.count()-1)
        
        wa.addItem("MOVE ABS")
        self.actions.add( motor.move_abs,
                         [self.inputs.pos_target.get, self.inputs.velocity.get], 
                         after=reset, feedback=self.feedback,
                         ).connect_item(wa, wa.count()-1)
                         
        wa.addItem("MOVE REL")
        self.actions.add( motor.move_rel,
                         [self.inputs.pos_target.get, self.inputs.velocity.get], 
                         after=reset, feedback=self.feedback,
                         ).connect_item(wa, wa.count()-1)
                         
        wa.addItem("MOVE VEL")
        self.actions.add( motor.move_vel,
                         [self.inputs.velocity.get], 
                         after=reset, feedback=self.feedback,
                         ).connect_item(wa, wa.count()-1)
        
         
        # Add posname in the list of actions
        list_action = [(name, motor.move_name, [name, self.inputs.velocity.get]) for name in motor.posnames]
        wa.insertSeparator(wa.count())
        
        for i,(name,func, inputs) in enumerate(list_action, start=wa.count()):
            wa.addItem(name)
            self.actions.add(func, inputs, after=reset, feedback=self.feedback).connect_item(wa, i)
                
record_widget_factory("line", "Motor", MotorLine)




from pydevmgr_core import NodeVar, BaseData
from pydevmgr_core_qt import find_ui, BaseUiLinker, get_style, record_widget_factory, Outputs
from pydantic import BaseModel
from .io import find_ui 

from PyQt5 import uic
from PyQt5.QtWidgets import QFrame



class CodeVal_O(Outputs.Base):
    """ widget output for a trio of (code, txt, group) 
    
    e.g. for a substate we get the substate code, its text representation and the group which will 
          be used to color the label. Note that group is a pydevmgr thing only.
    """
    def __init__(self, output, fmt="%s: %s"):
        if isinstance(fmt, str):
            fmt = lambda x,y, fmt=fmt: fmt%(x,y)
        
        if hasattr(output, "setText") and hasattr(output, "setStyleSheet") :                            
            def set_output(ctg): # code, text, group 
                c,t = ctg
                output.setText(fmt(c,t))
                # output.setStyleSheet(get_style(g))                       
        else:
            raise ValueError("Invalid input output combination")            
        super().__init__(set_output)
Outputs.Code = CodeVal_O 


class VltDeviceStatData(BaseModel):   
    error_code: NodeVar[int] = 0
    error_text: NodeVar[str] = ""
    
class VltDeviceData(BaseData):
    StatData = VltDeviceStatData # save StatData class here 
    
    stat: StatData = StatData()
    ignored: NodeVar[bool] = False
    name: str = "" # device name  
                
class VltDeviceCtrlUi(QFrame):
    def __init__(self,*args, **kwargs):
        super().__init__(*args, **kwargs)
        uic.loadUi(find_ui('device_ctrl_frame.ui'), self) 

    
class VltDeviceCtrl(BaseUiLinker):
    Widget = VltDeviceCtrlUi
    Data = VltDeviceData
    class Config(BaseUiLinker.Config):
        show_ignore_check_box: bool = True
    
    def init_vars(self):
        # these shall get stat data
        self.outputs.error    = self.outputs.Code(self.widget.error_txt )
                
        self.outputs.name =  self.outputs.Str(self.widget.name)
        
        self.outputs.feedback = self.outputs.Feedback(self.widget.feedback)
        self.inputs.ignored = self.inputs.NBool(self.widget.check)
                    
    def feedback(self, er, msg=''):
        self.outputs.feedback.set((er, msg))       
        
    def update(self, data : VltDeviceData) -> None:
        """ update the ui from the data structure """
        stat = data.stat
                                
        self.outputs.error.set( (stat.error_code, stat.error_text) )
        
        if self.config.show_ignore_check_box:
            if self.inputs.ignored.get() != data.ignored:             
                self.inputs.ignored.set_input(data.ignored)    
                
        
    def setup_ui(self, device, data):   
        super().setup_ui(device, data)
                                    
        self.outputs.name.set(data.name or device.key)
        
                
        self.setup_actions(device, data)
        if self.config.show_ignore_check_box:
            self.setup_ignore_checkbox(device,data)
        else:
            self.widget.check.hide()
            
    def setup_ignore_checkbox(self, device, data):
        w = self.widget.check
        w.setChecked(not device.ignored.get())                
        self.actions.add(device.ignored.set, [self.inputs.ignored.get]).connect_checkbox(w)
    
    def setup_actions(self, device, data):
        wa = self.widget.state_action
        wa.clear()
        wa.addItem("")
        
        action_list = [
         ("INIT",   device.init,   []), 
         ("ENABLE", device.enable, []),
         ("DISABLE",device.disable,[]),
         ("RESET",  device.reset,  []) 
        ]
        reset = lambda: wa.setCurrentIndex(0)
        
        for i, (name,func,inputs) in  enumerate(action_list, start=wa.count()):
            wa.addItem(name)
            action = self.actions.add(func, inputs, after=reset, feedback=self.feedback)
            action.connect_item(wa, i)
        
record_widget_factory("ctrl", "Vlt", VltDeviceCtrl)
